import asyncio 
import sys
import struct

from common.constants import *
from common.utils import *

# Handles multiple, concurrent commands to many servers at once
class MFSMultiConn:
	# Initializes class, sets timeout for a single task
	def __init__(self,timeout=5,addresses=[]):
		self.timeout = timeout
		self.addresses = addresses

	# Registers a server (host,port), to send a command to
	def register(self,host,port):
		self.addresses.append((host,port))

	# Checks if a given server (host,port) is registered for sending commands
	def is_registered(self,host,port):
		return (host,port) in self.addresses

	# Sends a given command asynchronously to all (previously) registered servers and applies the global timeout on all of them.
	# It is ensured that this function completes within give timeout
	# returns a dictionary hostkey->(datain, length) where hostkey is 'ip:port' (eg. '10.10.10.12:9422), 
	# if a given server can't be reached or gives no valid answer its hostkey will be absent from the dictionary, thus dictionary may be empty if no command completed successfully
	# postprocess is a function (lambda) that will be applied to each datain before storing it in the dictionary
	def command(self,cmdout,cmdin,dataout=None,postprocess=None):
		if (sys.version_info[0]==3 and sys.version_info[1]<7):
			loop = asyncio.get_event_loop()
			return loop.run_until_complete(self.async_command(cmdout,cmdin,dataout,postprocess))
		else:
			return asyncio.run(self.async_command(cmdout,cmdin,dataout,postprocess))

	# Internal: coordinates async tasks and collects results
	async def async_command(self,cmdout,cmdin,dataout,postprocess):
		results = []
		# Create a list of tasks
		if dataout and isinstance(dataout, list):
			tasks = [self.multiple_command(host, port, cmdout, cmdin, dataout) for host, port in self.addresses]
		else:
			tasks = [self.single_command(host, port, cmdout, cmdin, dataout, postprocess) for host, port in self.addresses]
		if not tasks:
			return dict(results)
		# Schedule tasks
		if (sys.version_info[0]==3 and sys.version_info[1]<7):
			loop = asyncio.get_event_loop()
			tasks = [loop.create_task(task) for task in tasks]
		else:
			tasks = [asyncio.create_task(task) for task in tasks]
		try:
			# Attempt to wait for all tasks with a timeout
			done, pending = await asyncio.wait(tasks, timeout=self.timeout)
			for task in done:
				if not task.cancelled() and not task.exception():
					results.append(task.result()) # Collect a single task output
		except asyncio.TimeoutError:
			raise Exception("Timeout during executing single_command")  # This line will not be hit using this structure.
		for task in pending:
			task.cancel() # Cancel pending tasks after timeout
			try:
				await task
			except asyncio.CancelledError:
				pass #do nothing about it
		return dict(results) # Convert list of tuples (successful task results) to dictionary
	
	# Internal: perform a single mfs i/o command 
	async def single_command(self,host,port,cmdout,cmdin,dataout,postprocess):
		if dataout:
			l = len(dataout)
			msg = struct.pack(">LL",cmdout,l) + dataout
		else:
			msg = struct.pack(">LL",cmdout,0)
		reader, writer = await asyncio.wait_for(asyncio.open_connection(host, port),timeout=2)
		cmdok = 0
		errcnt = 0
		error = None
		try: 
			while cmdok==0:
				badans = 0
				try:
					writer.write(msg)
					await writer.drain()
					while cmdok==0:
						header = await reader.readexactly(8)
						if header:
							cmd,length = struct.unpack(">LL",header)
							if cmd==cmdin:
								datain = await reader.readexactly(length)
								cmdok = 1
							elif cmd!=ANTOAN_NOP:
								badans = 1
								raise Exception
						else:
							raise Exception
				except Exception:
					if errcnt<3:
						writer.close()
						if not (sys.version_info[0]==3 and sys.version_info[1]<7):
							await writer.wait_closed()
						reader, writer = await asyncio.wait_for(asyncio.open_connection(host, port),timeout=3)
						errcnt+=1
					else:
						if badans:
							raise MFSCommunicationError("bad answer")
						else:
							raise MFSCommunicationError()
		except Exception as e:
			error = e
		finally: #clean up whatever happened
			writer.close()
			if not (sys.version_info[0]==3 and sys.version_info[1]<7):
				await writer.wait_closed()
		hostkey = "%s:%u" % (host,port)
		if postprocess: # apply postprocess function to datain
			return hostkey,postprocess(host,port,datain,length,error)
		else: # or return datain as is
			return hostkey,(datain,length)

	# Internal: perform synchronously multiple mfs i/o commands with a single peer (during a single connection) sending different data to each command
	# returns a list of answers (datain, length) for all commands
	async def multiple_command(self,host,port,cmdout,cmdin,dataout_arr):
		datain_arr,length_arr = [],[]
		reader, writer = await asyncio.wait_for(asyncio.open_connection(host, port),timeout=2)
		try:
			for dataout in dataout_arr:
				cmdok = 0
				errcnt = 0
				l = len(dataout)
				msg = struct.pack(">LL",cmdout,l) + dataout
				while cmdok==0:
					badans = 0
					try:
						writer.write(msg)
						await writer.drain()
						while cmdok==0:
							header = await reader.readexactly(8)
							if header:
								cmd,length = struct.unpack(">LL",header)
								if cmd==cmdin:
									datain = await reader.readexactly(length)
									cmdok = 1
								elif cmd!=ANTOAN_NOP:
									badans = 1
									raise Exception
							else:
								raise Exception
					except Exception:
						if errcnt<3:
							writer.close()
							if not (sys.version_info[0]==3 and sys.version_info[1]<7):
								await writer.wait_closed()
							reader, writer = await asyncio.wait_for(asyncio.open_connection(host, port),timeout=3)
							errcnt+=1
						else:
							if badans:
								raise MFSCommunicationError("bad answer")
							else:
								raise MFSCommunicationError()
				datain_arr.append(datain)
				length_arr.append(length)
		finally: #clean up whatever happened
			writer.close()
			if not (sys.version_info[0]==3 and sys.version_info[1]<7):
				await writer.wait_closed()
		hostkey = "%s:%u" % (host,port)
		return hostkey,(datain_arr,length_arr)


########################################################################################
# Handles a single connection for sending requests and commands to any MooseFS server
class MFSConn:
	def __init__(self,host,port):
		self.host = host
		self.port = port
		self.socket = None
		self.connect()

	def __del__(self):
		try:
			if self.socket:
				self.socket.close()
#				print "connection closed with: %s:%u" % (self.host,self.port)
			self.socket = None
		except AttributeError:
			pass

	def connect(self):
		cnt = 0
		while self.socket == None and cnt<3:
			self.socket = socket.socket()
			self.socket.settimeout(1)
			try:
				self.socket.connect((self.host,self.port))
			except Exception:
				self.socket.close()
				self.socket = None
				cnt += 1
		if self.socket==None:
			self.socket = socket.socket()
			self.socket.settimeout(1)
			self.socket.connect((self.host,self.port))
#		else:
#			print "connected to: %s:%u" % (self.host,self.port)

	def close(self):
		if self.socket:
			self.socket.close()
			self.socket = None

	def mysend(self,msg):
		if self.socket == None:
			self.connect()
		totalsent = 0
		while totalsent < len(msg):
			sent = self.socket.send(msg[totalsent:])
			if sent == 0:
				raise RuntimeError("socket connection broken")
			totalsent = totalsent + sent

	def myrecv(self,leng):
		msg = bytes(0)
		while len(msg) < leng:
			chunk = self.socket.recv(leng-len(msg))
			if len(chunk) == 0:
				raise RuntimeError("socket connection broken")
			msg = msg + chunk
		return msg
	
	def command(self,cmdout,cmdin,dataout=None):
		if dataout:
			l = len(dataout)
			msg = struct.pack(">LL",cmdout,l) + dataout
		else:
			msg = struct.pack(">LL",cmdout,0)
		cmdok = 0
		errcnt = 0
		while cmdok==0:
			badans = 0
			try:
				self.mysend(msg)
				while cmdok==0:
					header = self.myrecv(8)
					cmd,length = struct.unpack(">LL",header)
					if cmd==cmdin:
						datain = self.myrecv(length)
						cmdok = 1
					elif cmd!=ANTOAN_NOP:
						badans = 1
						raise Exception
			except Exception:
				if errcnt<3:
					self.close()
					self.connect()
					errcnt+=1
				else:
					if badans:
						raise MFSCommunicationError("bad answer")
					else:
						raise MFSCommunicationError()
		return datain,length

# Handles a single connection for sending requests and commands to Master server (leader or follower)
class MasterConn(MFSConn):
	def __init__(self,host,port):
		MFSConn.__init__(self,host,port)
		self.version = (0,0,0)
		self.pro = -1
		self.featuremask = 0

	def set_version(self,versionxyzp):
		self.version,self.pro = version_convert(versionxyzp)
		(_,self.sortver,self.pro) = version_str_sort_pro(versionxyzp)
		if self.version>=(3,0,72):
			self.featuremask |= (1<<FEATURE_EXPORT_UMASK)
		if (self.version>=(3,0,112) and self.version[0]==3) or self.version>=(4,21,0):
			self.featuremask |= (1<<FEATURE_EXPORT_DISABLES)
		if self.version>=(4,27,0):
			self.featuremask |= (1<<FEATURE_SESSION_STATS_28)
		if self.version>=(4,29,0):
			self.featuremask |= (1<<FEATURE_INSTANCE_NAME)
		if self.version>=(4,35,0):
			self.featuremask |= (1<<FEATURE_CSLIST_MODE)
		if self.version>=(4,44,0):
			self.featuremask |= (1<<FEATURE_SCLASS_IN_MATRIX)
		if self.version>=(4,51,0):
			self.featuremask |= (1<<FEATURE_DEFAULT_GRACEPERIOD)
		if self.version>=(4,53,0):
			self.featuremask |= (1<<FEATURE_LABELMODE_OVERRIDES)
		if self.version>=(4,57,0):
			self.featuremask |= (1<<FEATURE_SCLASSGROUPS)	

	def version_at_least(self,v1,v2,v3):
		return (self.version>=(v1,v2,v3))
	def version_less_than(self,v1,v2,v3):
		return (self.version<(v1,v2,v3))
	def version_is(self,v1,v2,v3):
		return (self.version==(v1,v2,v3))
	def version_unknown(self):
		return (self.version==(0,0,0))
	def is_pro(self):
		return self.pro
	def has_feature(self,featureid):
		return True if (self.featuremask & (1<<featureid)) else False
	
	def sort_ver(self):
		sortver = "%05u_%03u_%03u" % self.version
		if self.pro==1:
			sortver += "_2"
		elif self.pro==0:
			sortver += "_1"
		else:
			sortver += "_0"
		return sortver



from common.constants import *
from common.utils import *
from common.utilsgui import *

def render(dp, fields, vld, readonly, selectable):
	CSorder = fields.getint("CSorder", 0)
	CSrev = fields.getint("CSrev", 0)
	CScsid = fields.getstr("CScsid", "")
	CSlabel = fields.getstr("CSlabel", "ANY").upper()

	servers,dservers = dp.get_chunkservers_by_state(CSorder,CSrev)

	usedsum = 0
	totalsum = 0
	for cs in servers:
		if cs.total>0:
			usedsum+=cs.used
			totalsum+=cs.total
	if totalsum>0:
		avgpercent = (usedsum*100.0)/totalsum
	else:
		avgpercent = 0

	# sometimes cgi shows only a single server
	if (CScsid!="None" and CScsid!=""): 
		flt = lambda cs: ("%s:%s" % (cs.strip,cs.port)==CScsid)
		servers = list(filter(flt, servers))
		dservers = list(filter(flt, dservers))

	out = []
	if len(servers)>0:
		labels_dict = {}
		for cs in servers:
			for label in cs.labelstr.split(","):
				if label!="-":
					if label in labels_dict:
						labels_dict[label] += 1
					else:
						labels_dict[label] = 1
		if selectable and len(servers)>1 and len(labels_dict)>0:
			out.append("""<form action="#"><div class="tab_title">Chunkservers, label: """ )
			out.append("""<div class="select-fl"><select name="server" size="1" onchange="document.location.href='%s&CSlabel='+this.options[this.selectedIndex].value">""" % fields.createrawlink({"CSlabel":""}))
			entrystr = []
			entrydesc = {}
			entrystr.append("ANY")
			srv_count = len(servers)
			entrydesc["ANY"] =  "Any (%u server%s)" % (srv_count, "s" if srv_count>1 else "")
			labels_dict = {key: labels_dict[key] for key in sorted(labels_dict)}
			select_any = True
			for label in labels_dict:
				srv_count = labels_dict[label]
				label = label.upper()		
				entrystr.append(label)
				entrydesc[label] = "%s (%u server%s)" % (label, srv_count, "s" if srv_count>1 else "")
				if label==CSlabel:
					select_any = False
			if select_any:
				CSlabel = "ANY" # wrong label given in url, force selecting "ANY" label
			for estr in entrystr:
				if estr==CSlabel:
					out.append("""<option value="%s" selected="selected">%s</option>""" % (estr,entrydesc[estr]))
				else:
					out.append("""<option value="%s">%s</option>""" % (estr,entrydesc[estr]))
			out.append("""</select><span class="arrow"></span></div>""")
			out.append("""</div></form>""")
		else:
			if (len(servers)==1):
				out.append("""<div class="tab_title">Chunkserver details</div>""")
			else:
				out.append("""<div class="tab_title">Chunkservers</div>""")
		out.append("""<table class="acid_tab acid_tab_zebra_C1_C2 acid_tab_storageid_mfscs">""")
		out.append("""	<tr>""")
		out.append("""		<th rowspan="2" class="acid_tab_enumerate">#</th>""")
		out.append("""		<th rowspan="2">Id</th>""")
		out.append("""		<th rowspan="2">Host</th>""")
		out.append("""		<th rowspan="2">IP</th>""")
		out.append("""		<th rowspan="2">Port</th>""")
		out.append("""		<th rowspan="2">Labels</th>""")
		out.append("""		<th rowspan="2">Version</th>""")
		out.append("""		<th rowspan="2">Queue</th>""")
		out.append("""		<th rowspan="2">Queue state</th>""")
		out.append("""		<th rowspan="2">Maintenance</th>""")
		out.append("""		<th colspan="4">'Regular' hdd space</th>""")
		if dp.master().version_at_least(3,0,38):
			out.append("""		<th colspan="5">'Marked for removal' hdd space</th>""")
		else:
			out.append("""		<th colspan="4">'Marked for removal' hdd space</th>""")
		out.append("""	</tr>""")
		out.append("""	<tr>""")
		out.append("""		<th>chunks</th>""")
		out.append("""		<th>used</th>""")
		out.append("""		<th>total</th>""")
		out.append("""		<th class="progbar">% used</th>""")
		if dp.master().version_at_least(3,0,38):
			out.append("""		<th>status</th>""")
		out.append("""		<th>chunks</th>""")
		out.append("""		<th>used</th>""")
		out.append("""		<th>total</th>""")
		out.append("""		<th class="progbar">% used</th>""")
		out.append("""	</tr>""")

	if (CSlabel!="ANY" and CSlabel!=""): 
		flt = lambda cs: (CSlabel in cs.labelstr)
		servers = list(filter(flt, servers))
		dservers = list(filter(flt, dservers))

	# iterate all connected servers
	for cs in servers:
		if cs.strip!=cs.stroip:
			cs.strip = "%s&#8239;&rightarrow;&#8239;%s" % (cs.stroip,cs.strip)
		if (cs.is_maintenance_off()):
			mmchecked = ""
			mmstr = "OFF"
			mmurl = fields.createhtmllink({"CSmaintenanceon":("%s:%u" % (cs.stroip,cs.port))})
			mmicon = ''
		elif (cs.is_maintenance_on()):
			mmchecked = "checked"
			mmstr = "ON"
			mmurl = fields.createhtmllink({"CSmaintenanceoff":("%s:%u" % (cs.stroip,cs.port))})
			mmicon = html_icon('icon-wrench','cs_maintain')
		else:
			mmchecked = "checked"
			mmstr = "TMP"
			mmurl = fields.createhtmllink({"CSmaintenanceoff":("%s:%u" % (cs.stroip,cs.port))})
			mmicon = html_icon('icon-wrench','cs_maintain_tmp')
		out.append("""	<tr>""")
		out.append("""		<td align="right"></td>""")
		out.append("""		<td align="center">%u</td>""" % cs.csid)
		# hostname
		out.append("""		<td align="left"><span class="text-icon">%s%s</span></td>""" % (cs.host, mmicon))
		out.append("""		<td align="center"><span class="sortkey">%s </span>%s</td>""" % (cs.sortip,cs.strip))
		out.append("""		<td align="center">%u</td>""" % cs.port)
		# labels
		out.append("""		<td align="left" class="monospace">%s</td>""" % cs.labelstr)
		issues = vld.check_cs_version(cs)
		out.append("""		<td align="center"><span class="sortkey">%s </span><span class="%s">%s</span></td>""" % (cs.sortver,'',issues.span(cs.strver.replace("PRO","<small>PRO</small>"))))
		# queue 
		out.append("""		<td align="right">%u</td>""" % (cs.queue))
		if cs.queue_state==CS_LOAD_FAST_REBALANCE:
			out.append("""		<td align="center"><a style="cursor:default" title="%s" href="%s">%s</a></td>""" % (cs.queue_state_info,fields.createhtmllink({"CSbacktowork":("%s:%u" % (cs.strip,cs.port))}),cs.queue_state_str))
		else:
			issues = vld.check_cs_queue_state(cs)
			info = cs.queue_state_info if cs.queue_state!=CS_LOAD_NORMAL else ""
			out.append("""		<td align="center"><span data-tt-info="%s">%s</span></td>""" % (info, issues.span(cs.queue_state_str)))
		# maintenance
		if dp.cluster.leaderfound():
			if readonly:
				out.append("""		<td class="center">%s</td>""" % mmstr)
			else:
				out.append("""		<td class="center">""")
				out.append("""			<span><label class="switch width34" for="maintenance-checkbox-%s"><input type="checkbox" id="maintenance-checkbox-%s" onchange="window.location.href='%s'" %s/><span class="slider round checked_blue"><span class="slider-text">%s</span></span></label></span>""" % (cs.hostkey, cs.hostkey, mmurl, mmchecked,mmstr))
				out.append("""		</td>""")
		else:
			out.append("""		<td class="center">n/a</td>""")
		# chunks
		out.append("""		<td align="right">%s</td>""" % (decimal_number_html(cs.chunks)))
		issues = vld.check_cs_hdds(cs)
		out.append("""		<td align="right"><span class="sortkey">%u </span><span data-tt-info="%s B">%s</span></td>""" % (cs.used, decimal_number(cs.used), issues.span(humanize_number(cs.used, "&nbsp;"))))
		out.append("""		<td align="right"><span class="sortkey">%u </span><span data-tt-info="%s B">%s</span></td>""" % (cs.total,decimal_number(cs.total),humanize_number(cs.total,"&nbsp;")))
		if cs.total>0:
			usedpercent = (cs.used*100.0)/cs.total
			if avgpercent-usedpercent>=0.1:
				diffstr = "%.1f pp. lower than all chunkservers' average (%.1f%%)" % (avgpercent-usedpercent, avgpercent)
			elif usedpercent-avgpercent>=0.1:
				diffstr = "%.1f pp. higher than all chunkservers' average (%.1f%%)" % (usedpercent-avgpercent, avgpercent)
			else:
				diffstr = "equal to chunkservers' average"
			out.append("""		<td align="center"><span class="sortkey">%.10f </span><div class="PROGBOX"><div class="PROGCOVER" style="width:%.2f%%;"></div><div class="PROGAVG" style="width:%.2f%%"></div><div class="PROGVALUE"><span data-tt-info="%s">%.1f</span></div></div></td>""" % (usedpercent,100.0-usedpercent,avgpercent,diffstr,usedpercent))
		else:
			out.append("""		<td align="center"><span class="sortkey">-1 </span><div class="PROGBOX"><div class="PROGCOVER" style="width:100%;"></div><div class="PROGVALUE"><span></span></div></div></td>""")
		if dp.master().version_at_least(3,0,38):

			if cs.tdchunks==0 or dp.cluster.leaderfound()==0:
				out.append("""		<td align="center">-</td>""")
			else:
				issues = vld.check_cs_mfr_status(cs)
				clz = "OK" if issues.len()==0 else "MFRREADY" if cs.mfrstatus==MFRSTATUS_READY else ""
				out.append("""		<td>%s</td>""" % issues.span(cs.get_mfr_status_str(),clz))
		out.append("""		<td align="right">%s</td><td align="right"><span class="sortkey">%u </span><span data-tt-info="%s B">%s</span></td><td align="right"><span class="sortkey">%u </span><span data-tt-info="%s B">%s</span></td>""" % (decimal_number_html(cs.tdchunks),cs.tdused,decimal_number(cs.tdused),humanize_number(cs.tdused,"&nbsp;"),cs.tdtotal,decimal_number(cs.tdtotal),humanize_number(cs.tdtotal,"&nbsp;")))
		if (cs.tdtotal>0):
			usedpercent = (cs.tdused*100.0)/cs.tdtotal
			out.append("""		<td align="center"><span class="sortkey">%.10f </span><div class="PROGBOX"><div class="PROGCOVER" style="width:%.2f%%;"></div><div class="PROGVALUE"><span>%.1f</span></div></div></td>""" % (usedpercent,100.0-usedpercent,usedpercent))
		else:
			out.append("""		<td align="center"><span class="sortkey">-1 </span><div class="PROGBOX"><div class="PROGCOVER" style="width:100%;"></div><div class="PROGVALUE"><span></span></div></div></td>""")
		out.append("""	</tr>""")
	out.append("""</table>""")

	if len(dservers)>0:
		if (len(dservers)==1):
			out.append("""<div class="tab_title">Disconnected chunkserver details</div>""")
		else:
			out.append("""<div class="tab_title">Disconnected chunkservers</div>""")
		out.append("""<table class="acid_tab acid_tab_zebra_C1_C2 acid_tab_storageid_mfsdiscs">""")
		out.append("""	<tr>""")
		out.append("""		<th class="acid_tab_enumerate">#</th>""")
		out.append("""		<th>Host</th>""")
		out.append("""		<th>IP</th>""")
		out.append("""		<th>Port</th>""")
		out.append("""		<th>Id</th>""")
		out.append("""		<th>Maintenance</th>""")
		if (not readonly):
			if dp.cluster.leaderfound() and dp.cluster.deputyfound()==0:
				out.append("""		<th class="acid_tab_skip">Remove</th>""")
			else:
				out.append("""		<th class="acid_tab_skip">Temporarily remove</th>""")
		out.append("""	</tr>""")

	# iterate all disconnected servers
	for cs in dservers:
		if cs.strip!=cs.stroip:
			cs.strip = "%s &rightarrow; %s" % (cs.stroip,cs.strip)
		out.append("""	<tr>""")
		if dp.cluster.leaderfound()==0:
			mmicon = ''
			cl = "DISCONNECTED"
		elif cs.is_maintenance_off():
			mmchecked = ""
			mmstr = "OFF"
			mmurl = fields.createhtmllink({"CSmaintenanceon":("%s:%u" % (cs.stroip,cs.port))})
			mmicon = html_icon('icon-error','cs_unreachable')
			cl = "DISCONNECTED"
		elif cs.is_maintenance_on():
			mmchecked = "checked"
			mmstr = "ON"
			mmurl = fields.createhtmllink({"CSmaintenanceoff":("%s:%u" % (cs.stroip,cs.port))})
			mmicon = html_icon('icon-warning','cs_unreachable_maintain')+html_icon('icon-wrench','cs_unreachable_maintain')
			cl = "MAINTAINED"
		else:
			if cs.maintenanceto==0xFFFFFFFF:
				mmstr = "ON" #should not be here
			else:
				mmstr = """%u:%02u""" % (cs.maintenanceto // 60, cs.maintenanceto % 60)
			mmchecked = "checked"
			mmurl = fields.createhtmllink({"CSmaintenanceoff":("%s:%u" % (cs.stroip,cs.port))})
			mmicon = html_icon('icon-warning','cs_maintain_tmp')+html_icon('icon-wrench','cs_maintain_tmp')+html_icon('icon-stopwatch','cs_maintain_tmp', 1, 'icon-blue')
			cl = "TMPMAINTAINED"
		out.append("""		<td align="right"></td><td align="left"><span class="%s text-icon">%s%s</span></td>""" % (cl,cs.host,mmicon))
		out.append("""		<td align="center"><span class="sortkey">%s </span><span class="%s">%s</span></td>""" % (cs.sortip,cl,cs.strip))
		out.append("""		<td align="center"><span class="%s">%u</span></td>""" % (cl,cs.port))
		out.append("""		<td align="right"><span class="%s">%u</span></td>""" % (cl,cs.csid))
		if dp.cluster.leaderfound(): #maintenance
			if readonly:
				out.append("""		<td class="center">%s</td>""" % mmstr)
			else:
				out.append("""		<td class="center">""")
				out.append("""			<span><label class="switch width38" for="maintenance-checkbox-%s"><input type="checkbox" id="maintenance-checkbox-%s" onchange="window.location.href='%s'" %s/><span class="slider round checked_blue"><span class="slider-text countdown">%s</span></span></label></span>""" % (cs.hostkey, cs.hostkey, mmurl, mmchecked, mmstr))
				out.append("""		</td>""")
		else:
			out.append("""		<td align="center"><span class="%s">not available</td>""" % cl)
		if (not readonly): #remove cmd
			if dp.cluster.leaderfound() and dp.cluster.deputyfound()==0:
				out.append("""		<td align="center"><a class="VISIBLELINK" href="%s">click to remove</a></td>""" % (fields.createhtmllink({"CSremove":("%s:%u" % (cs.stroip,cs.port))})))
			elif dp.master().version_at_least(3,0,67):
				out.append("""		<td align="center"><a class="VISIBLELINK" href="%s">click to temporarily remove</a></td>""" % (fields.createhtmllink({"CStmpremove":("%s:%u" % (cs.stroip,cs.port))})))
			else:
				out.append("""		<td align="center">not available</td>""")

		out.append("""	</tr>""")
	out.append("""</table>""")
	
	return out



from common.constants import *
from common.utils import *
from common.conn import *
from common.models import *

# # called asynchronously after receiving the answer to the info command from each master server
# # it should decode the answer and built the master server object
# def process_info_answer(host,port,data,length,error):
# 	# print("process_info_answer data: %s\n\n" % data)
# 	version = (0,0,0)
# 	workingstate = STATE_DUMMY
# 	statestr = ""
# 	statecolor = 1
# 	memusage = 0
# 	syscpu = 0
# 	usercpu = 0
# 	lastsuccessfulstore = 0
# 	lastsaveseconds = 0
# 	lastsavestatus = 0
# 	metaversion = 0
# 	exportschecksum = None
# 	metaid = None
# 	lastsavemetaversion = None
# 	lastsavemetachecksum = None
# 	usectime = None
# 	chlogtime = 0
# 	try:
# 		if length==52:
# 			version = (1,4,0)
# 		elif length==60:
# 			version = (1,5,0)
# 		elif length==68 or length==76 or length==101: #(size < 121 => version < 2.0.0)
# 			version = struct.unpack(">HBB",data[:4])
# 		# (size = 121,version >= 2.0.0) (size = 129,version >= 3.0.32) (size = 173,version >= 4.4.0)
# 		elif length==121 or length==129 or length==137 or length==149 or length==173 or length==181 or length==193 or length==205:
# 			offset = 8 if (length>=137 and length!=173) else 0
# 			version = struct.unpack(">HBB",data[:4])
# 			memusage,syscpu,usercpu = struct.unpack(">QQQ",data[4:28])
# 			syscpu/=10000000.0
# 			usercpu/=10000000.0
# 			if length==205:
# 				offset += 4
# 			lastsuccessfulstore,lastsaveseconds,lastsavestatus = struct.unpack(">LLB",data[offset+92:offset+101])
# 			if version>(2,0,14):
# 				lastsaveseconds = lastsaveseconds / 1000.0
# 			workingstate,nextstate,stablestate,sync,leaderip,changetime,metaversion = struct.unpack(">BBBBLLQ",data[offset+101:offset+121])
# 			if length>=129:
# 				exportschecksum = struct.unpack(">Q",data[offset+121:offset+129])[0]
# 			if length>=173:
# 				metaid,lastsavemetaversion,lastsavemetachecksum = struct.unpack(">QQL",data[offset+129:offset+149])
# 			if length==149 or length==193 or length==205:
# 				usectime,chlogtime = struct.unpack(">QL",data[length-12:length])
# 			if workingstate==STATE_MASTERCE and nextstate==STATE_MASTERCE and stablestate==STATE_MASTERCE and sync==STATE_MASTERCE:
# 				statestr = state_name(workingstate) # "MASTER"
# 				statecolor = 0
# 			elif stablestate==0 or workingstate!=nextstate:
# 				statestr = "transition %s -> %s" % (state_name(workingstate),state_name(nextstate))
# 				statecolor = 8
# 			else:
# 				statestr = state_name(workingstate)
# 				statecolor = state_color(workingstate,sync)
# 		else:
# 			version = (0,0,0) # unknown version
# 		# except Exception:
# 		# 	statestr = STATE_STR_BUSY
# 		# 	statecolor = 7
# 	except Exception:
# 		workingstate = STATE_UNREACHABLE
# 		statestr = STATE_STR_UNREACHABLE
# 	ms = MasterServer(host,port,version,workingstate,sync,statestr,statecolor,metaversion,memusage,syscpu,usercpu,lastsuccessfulstore,lastsaveseconds,lastsavestatus,exportschecksum,metaid,lastsavemetaversion,lastsavemetachecksum,usectime,chlogtime)
# 	return ms

# Represents state of master servers in the cluster: leader, elect etc.
class Cluster:
	# possible_leader_ip is a hint for possible leading master IP address (the previously known leader)
	def __init__(self, masterhost, masterport, possible_leader_ip=None):
		self.masterhost = masterhost
		self.masterport = masterport

		# Don't use the following underscored variables directly, use the methods instead
		self._masterconn = None
		self._masterlist = None
		self._masterinfo = None
		self._master_exportschecksum = 0
		self._master_metaid = 0
		self._leader_usectime = None

		self._errormsg = "Unknown error"

		self._leaderfound = 0
		self._electfound = 0
		self._usurperfound = 0
		self._followerfound = 0
		self._deputyfound = 0

		# Resolve master hostname to all IP addresses
		self.addresses = []
		for mhost in self.masterhost.replace(';',' ').replace(',',' ').split():
			try:
				for i in socket.getaddrinfo(mhost,self.masterport,socket.AF_INET,socket.SOCK_STREAM,socket.SOL_TCP):
					if i[0]==socket.AF_INET and i[1]==socket.SOCK_STREAM and i[2]==socket.SOL_TCP:
						self.addresses.append(i[4])
						# print("Master address: %s:%u" % i[4])
			except Exception:
				pass

		if len(self.addresses)==0:
			self._errormsg = """Can't resolve the MooseFS Master hostname (%s)""" % (masterhost)
		elif len(self.addresses)==1:
			self._errormsg = """Can't connect to the MooseFS Master server (%s)""" % (masterhost)
		else:
			self._errormsg = """Can't connect to MooseFS Master servers (%s)""" % (masterhost)

		# self.find_all_masters()
		
		if possible_leader_ip:
			# put possible leader to the beginning of the list
			self.addresses.sort(key=lambda x: x[0]!=possible_leader_ip)

		# Find master servers in the cluster, find them all if there is no hint for possible leading master IP address
		self.find_masters(find_all=(possible_leader_ip==None))

	# Returns True if any master server is found in the cluster
	def anyfound(self):
		return self._leaderfound or self._electfound or self._usurperfound or self._followerfound or self._deputyfound
	
	# Returns True if any leading master server is found in the cluster (leader, elect, usurper, or deputy)
	def leadingfound(self):
		return self._leaderfound or self._electfound or self._usurperfound or self._deputyfound

	# Return the current leading master server (connection)
	def master(self):
		return self._masterconn

	def errormsg(self):
		return self._errormsg

	def leaderfound(self):
		return self._leaderfound
	def followerfound(self):
		return self._followerfound
	def electfound(self):
		return self._electfound
	def usurperfound(self):
		return self._usurperfound
	def deputyfound(self):
		return self._deputyfound

	def masterinfo(self):
		return self._masterinfo
	def master_metaid(self):
		return self._master_metaid
	def master_exportschecksum(self):
		return self._master_exportschecksum


	# def find_all_masters(self):
	# 	multiconn = MFSMultiConn(0.5,self.addresses)
	# 	answers=multiconn.command(CLTOMA_INFO,MATOCL_INFO,None,process_info_answer) 



	def find_masters(self, find_all):
		# Don't use the following variables directly, use the methods instead
		self._masterconn = None
		self._masterlist = None
		self._masterinfo = None
		self._master_exportschecksum = 0
		self._master_metaid = 0
		self._leader_usectime = None

		self._leaderfound = 0
		self._electfound = 0
		self._usurperfound = 0
		self._followerfound = 0
		self._deputyfound = 0

		leaderinfo = None
		leader_exportschecksum = None
		leader_metaid = None
		leaderconn = None

		electinfo = None
		elect_exportschecksum = None
		elect_metaid = None
		electconn = None

		usurperinfo = None
		usurper_exportschecksum = None
		usurper_metaid = None
		usurperconn = None

		followerinfo = None
		follower_exportschecksum = None
		follower_metaid = None
		followerconn = None

		if find_all:
			self._masterlist = []

		# find leader
		for mhost,mport in self.addresses:
			conn = None
			version = (0,0,0)
			workingstate = STATE_DUMMY
			sync = 0
			statestr = "???"
			statecolor = 1
			memusage = 0
			syscpu = 0
			usercpu = 0
			lastsuccessfulstore = 0
			lastsaveseconds = 0
			lastsavestatus = 0
			metaversion = 0
			exportschecksum = None
			metaid = None
			lastsavemetaversion = None
			lastsavemetachecksum = None
			usectime = None
			chlogtime = 0
			try:
				conn = MasterConn(mhost,mport)
				try:
					data,length = conn.command(CLTOMA_INFO,MATOCL_INFO)
					if length==52:
						version = (1,4,0)
						conn.set_version(version)
						if self._leaderfound==0:
							leaderconn = conn
							leaderinfo = data
							self._leaderfound = 1
						statestr = "OLD MASTER (LEADER ONLY)"
						statecolor = 0
					elif length==60:
						version = (1,5,0)
						conn.set_version(version)
						if self._leaderfound==0:
							leaderconn = conn
							leaderinfo = data
							self._leaderfound = 1
						statestr = "OLD MASTER (LEADER ONLY)"
						statecolor = 0
					elif length==68 or length==76 or length==101:
						version = struct.unpack(">HBB",data[:4])
						conn.set_version(version)
						if self._leaderfound==0 and version<(1,7,0):
							leaderconn = conn
							leaderinfo = data
							self._leaderfound = 1
						if length==76:
							memusage = struct.unpack(">Q",data[4:12])[0]
						if length==101:
							memusage,syscpu,usercpu = struct.unpack(">QQQ",data[4:28])
							syscpu/=10000000.0
							usercpu/=10000000.0
							lastsuccessfulstore,lastsaveseconds,lastsavestatus = struct.unpack(">LLB",data[92:101])
						if version<(1,7,0):
							statestr = "OLD MASTER (LEADER ONLY)"
							statecolor = 0
						else:
							statestr = "UPGRADE THIS UNIT!"
							statecolor = 2
					elif length==121 or length==129 or length==137 or length==149 or length==173 or length==181 or length==193 or length==205:
						offset = 8 if (length>=137 and length!=173) else 0
						version = struct.unpack(">HBB",data[:4])
						conn.set_version(version)
						memusage,syscpu,usercpu = struct.unpack(">QQQ",data[4:28])
						syscpu/=10000000.0
						usercpu/=10000000.0
						if length==205:
							offset += 4
						lastsuccessfulstore,lastsaveseconds,lastsavestatus = struct.unpack(">LLB",data[offset+92:offset+101])
						if conn.version_at_least(2,0,14):
							lastsaveseconds = lastsaveseconds / 1000.0
						workingstate,nextstate,stablestate,sync,leaderip,changetime,metaversion = struct.unpack(">BBBBLLQ",data[offset+101:offset+121])
						if length>=129:
							exportschecksum = struct.unpack(">Q",data[offset+121:offset+129])[0]
						if length>=173:
							metaid,lastsavemetaversion,lastsavemetachecksum = struct.unpack(">QQL",data[offset+129:offset+149])
						if length==149 or length==193 or length==205:
							usectime,chlogtime = struct.unpack(">QL",data[length-12:length])
						if workingstate==STATE_MASTERCE and nextstate==STATE_MASTERCE and stablestate==STATE_MASTERCE and sync==0xFF:
							if self._leaderfound==0:
								leaderconn = conn
								leaderinfo = data
								self._leaderfound = 1
								leader_exportschecksum = exportschecksum
								leader_metaid = metaid
								self._leader_usectime = usectime
							statestr = state_name(workingstate) # "MASTER"
							statecolor = 0
						elif stablestate==0 or workingstate!=nextstate:
							statestr = "transition %s -> %s" % (state_name(workingstate),state_name(nextstate))
							statecolor = 8
						else:
							statestr = state_name(workingstate)
							statecolor = state_color(workingstate,sync)
							if workingstate==STATE_FOLLOWER:
								# if sync==0:
								# 	statestr += " (DESYNC)"
								# if sync==2:
								# 	statestr += " (DELAYED)"
								# if sync==3:
								# 	statestr += " (INIT)"
								self._followerfound = 1
								followerconn = conn
								followerinfo = data
								follower_exportschecksum = exportschecksum
								follower_metaid = metaid
							if workingstate==STATE_USURPER and self._usurperfound==0:
								self._usurperfound = 1
								usurperconn = conn
								usurperinfo = data
								usurper_exportschecksum = exportschecksum
								usurper_metaid = metaid
							if workingstate==STATE_ELECT and self._electfound==0:
								self._electfound = 1
								electconn = conn
								electinfo = data
								elect_exportschecksum = exportschecksum
								elect_metaid = metaid
							if (workingstate==STATE_LEADER or workingstate==STATE_DEPUTY) and self._leaderfound==0:
								leaderconn = conn
								leaderinfo = data
								self._leaderfound = 1
								if (workingstate==STATE_DEPUTY):
									self._deputyfound = 1
								leader_exportschecksum = exportschecksum
								leader_metaid = metaid
								self._leader_usectime = usectime
					else:
						if len(self.addresses)==1:
							self._errormsg = """Got wrong answer from the MooseFS Master server (%s) - likely the master version is too new""" % (self.masterhost)
						else:
							self._errormsg = """Got wrong answer from MooseFS Master servers (%s) - likely the masters version is too new""" % (self.masterhost)
						statestr = STATE_STR_WRONGANSWER
						statecolor = 1
				except Exception:
					statestr = STATE_STR_BUSY
					statecolor = 7
			except Exception:
				workingstate = STATE_UNREACHABLE
				statestr = STATE_STR_UNREACHABLE

			if conn and conn!=leaderconn and conn!=electconn and conn!=usurperconn and conn!=followerconn:
				del conn
			if find_all:
				self._masterlist.append(MasterServer(mhost,mport,version,workingstate,sync,statestr,statecolor,metaversion,memusage,syscpu,usercpu,lastsuccessfulstore,lastsaveseconds,lastsavestatus,exportschecksum,metaid,lastsavemetaversion,lastsavemetachecksum,usectime,chlogtime))

			if self._leaderfound:
				self._masterconn = leaderconn
				self._masterinfo = leaderinfo
				self._master_exportschecksum = leader_exportschecksum
				self._master_metaid = leader_metaid
				if not find_all:
					break # don't need to find other masters if leader is found
			elif self._electfound:
				self._masterconn = electconn
				self._masterinfo = electinfo
				self._master_exportschecksum = elect_exportschecksum
				self._master_metaid = elect_metaid
			elif self._usurperfound:
				self._masterconn = usurperconn
				self._masterinfo = usurperinfo
				self._master_exportschecksum = usurper_exportschecksum
				self._master_metaid = usurper_metaid
			elif self._followerfound:
				self._masterconn = followerconn
				self._masterinfo = followerinfo
				self._master_exportschecksum = follower_exportschecksum
				self._master_metaid = follower_metaid



	# Returns a list of all master servers in the cluster along with its highest metaversion and checksum (tuple)
	# Sorts if IMorder - sort order is provided, IMrev - reverse order
	def get_masterservers(self, IMorder=0, IMrev=False):
		if self._masterlist==None: # _masterlist could not be initialized if only the leader was found on initialization
			self.find_masters(True)
			# self.find_all_masters(None, True)

		if IMorder==None:
			return self._masterlist
		masterservers = []
		for ms in self._masterlist:
			if   IMorder==0:	key=lambda ms: (ms.sortip, ms.port)
			elif IMorder==1:	key=lambda ms: (ms.sortip)
			elif IMorder==2:	key=lambda ms: (ms.sortver)
			elif IMorder==3:	key=lambda ms: (ms.statecolor)
			elif IMorder==4:	key=lambda ms: (ms.usectime if ms.usectime!=None else 0)
			elif IMorder==5:	key=lambda ms: (ms.metaversion)
			elif IMorder==6:	key=lambda ms: (ms.metaid)
			elif IMorder==7:	key=lambda ms: (ms.chlogtime if ms.chlogtime!=None else 0)
			elif IMorder==8:	key=lambda ms: (ms.memusage)
			elif IMorder==9:	key=lambda ms: (ms.syscpu+ms.usercpu)
			elif IMorder==10:	key=lambda ms: (ms.lastsuccessfulstore)
			elif IMorder==11:	key=lambda ms: (ms.lastsaveseconds)
			elif IMorder==12:	key=lambda ms: (ms.lastsavestatus)
			elif IMorder==13:	key=lambda ms: (ms.lastsavemetaversion)
			elif IMorder==14:	key=lambda ms: (ms.lastsavemetachecksum)
			elif IMorder==15:	key=lambda ms: (ms.exportschecksum)
			else:				      key=lambda ms: (0)
			masterservers.append(ms)
		masterservers.sort(key=key)
		if IMrev:
			masterservers.reverse()
		return masterservers

	# Returns a list of all master servers in the cluster along with secdelta and metadelay
	# Sorts if IMorder - sort order is provided, IMrev - reverse order
	def update_masterservers_delays(self, IMorder=0, IMrev=False):
		mservers = self.get_masterservers(IMorder, IMrev)
		# Calculate necessary metrics from all metadata servers
		highest_saved_metaversion = 0
		highest_metaversion_checksum = 0
		master_minusectime = None
		master_maxusectime = None
		master_deltausectime = None
		for ms in mservers:
			if ms.lastsavemetaversion!=None and ms.lastsavemetachecksum!=None:
				if ms.lastsavemetaversion>highest_saved_metaversion:
					highest_saved_metaversion = ms.lastsavemetaversion
					highest_metaversion_checksum = ms.lastsavemetachecksum
				elif ms.lastsavemetaversion==highest_saved_metaversion:
					highest_metaversion_checksum |= ms.lastsavemetachecksum
			if self._leader_usectime==None or self._leader_usectime==0:
				if ms.usectime!=None and ms.usectime>0:
					if master_minusectime==None or ms.usectime<master_minusectime:
						master_minusectime = ms.usectime
					if master_maxusectime==None or ms.usectime>master_maxusectime:
						master_maxusectime = ms.usectime
				if master_maxusectime and master_minusectime:
					master_deltausectime = master_maxusectime - master_minusectime
		updated_masterlist = []
		for ms in mservers:
			secdelta = 0.0
			if ms.usectime!=None and ms.usectime!=0:
				if self._leader_usectime==None or self._leader_usectime==0:
					if master_deltausectime!=None:
						secdelta = (master_deltausectime) / 1000000.0
				else:
					if self._leader_usectime > ms.usectime:
						secdelta = (self._leader_usectime - ms.usectime) / 1000000.0
					else:
						secdelta = (ms.usectime - self._leader_usectime) / 1000000.0
			if ms.chlogtime==None or ms.chlogtime==0 or self._leader_usectime==None or self._leader_usectime==0:
				metadelay = None
			else:
				metadelay = self._leader_usectime/1000000.0 - ms.chlogtime
				if metadelay>1.0:
					metadelay-=1.0
				else:
					metadelay=0.0
			ms.secdelta = secdelta
			ms.metadelay = metadelay
			updated_masterlist.append(ms)
		self._masterlist = updated_masterlist
		return highest_saved_metaversion, highest_metaversion_checksum



from common.utilsgui import *

# renders forms with errors and prompts when leader master server is missing
def render(dp, fields, vld):
	cl = dp.cluster
	out = []
	out.append("""<div class="tab_title ERROR">Oops!</div>""")
	out.append("""<table class="FR MESSAGE">""")
	if not cl.anyfound():
		out.append("""	<tr>""")
		out.append("""		<td align="center">""")
		out.append("""			<span class="ERROR">Can't find masters (resolve given name)!</span><br />""")
		out.append("""			<form method="GET">""")
		out.append("""				Input your DNS master name: <input type="text" name="masterhost" value="%s" size="100">""" % htmlentities(cl.masterhost))
		for i in fields.createinputs(["masterhost"]):
			out.append("""				%s""" % (i))
		out.append("""				<input type="submit" value="Try it!">""")
		out.append("""			</form>""")
		out.append("""		</td>""")
		out.append("""	</tr>""")
	else:
		out.append("""	<tr>""")
		out.append("""		<td align="center">""")
		out.append("""			<span class="ERROR">Can't find working masters!</span><br />""")
		out.append("""			<form method="GET">""")
		out.append("""				Input your DNS master name: <input type="text" name="masterhost" value="%s" size="100"><br />""" % htmlentities(cl.masterhost))
		out.append("""				Input your master-client port number: <input type="text" name="masterport" value="%u" size="5"><br />""" % (cl.masterport))
		for i in fields.createinputs(["masterhost","masterport"]):
			out.append("""				%s""" % (i))
		out.append("""				<input type="submit" value="Try it!">""")
		out.append("""			</form>""")
		out.append("""		</td>""")
		out.append("""	</tr>""")
	out.append("""</table>""")
	return out


from common.constants import *
from common.utils import *
from common.utilsgui import *

def render(dp, fields, vld):
	MBorder = fields.getint("MBorder", 0)
	MBrev = fields.getint("MBrev", 0)

	out = []
	# out.append("""<div class="tab_title">Metadata backup loggers</div>""")
	out.append("""<table class="acid_tab acid_tab_zebra_C1_C2 acid_tab_storageid_mfsmbl">""")
	out.append("""	<tr>""")
	out.append("""		<th class="acid_tab_enumerate">#</th>""")
	out.append("""		<th>IP</th>""")
	out.append("""		<th>Version</th>""")
	out.append("""		<th style="border-right:0;">Metalogger host</th>""")
	out.append("""		<th></th>""")
	out.append("""	</tr>""")

	for ml in dp.get_metaloggers(MBorder, MBrev):
		issues = vld.check_ml_version(ml)
		out.append("""	<tr>""")
		# out.append("""		<td align="right"></td><td align="left">%s</td><td align="center"><span class="sortkey">%s </span>%s</td><td align="center"><span class="sortkey">%s </span>%s</td>""" % (ml.host,ml.sortip,ml.strip,ml.sortver,issues.span(ml.strver.replace("PRO","<small>PRO</small>"))))
		out.append("""    <td align="right"></td>""")
		out.append("""		<td align="center" style="width:6%%;"><span class="sortkey">%s </span>%s</td>""" % (ml.sortip, ml.strip))
		out.append("""		<td align="center" style="width:6%%;"><span class="sortkey">%s </span>%s</td>""" % (ml.sortver, issues.span(ml.strver.replace("PRO","<small>PRO</small>"))))
		out.append("""		<td align="center" style="width:16%%;border-right:0;">%s</td>""" % ml.host) 
		out.append("""		<td></td>""") 
		out.append("""	</tr>""")

	out.append("""</table>""")

	return out


from common.constants import *
from common.utils import *
from common.utilsgui import *
from common.models import *

def render(dp, fields, vld):
	ci = dp.get_clusterinfo()
	out = []
	out.append("""<div class="tab_title">Cluster summary</div>""")
	out.append("""<table class="FR no-hover">""")

	out.append("""	<tr>""")
	out.append("""		<th colspan="3">Cluster space</th>""")
	out.append("""		<th colspan="2">Trash</th>""")
	out.append("""		<th colspan="2">Sustained</th>""")
	out.append("""		<th colspan="3">File system objects</th>""")
	out.append("""		<th rowspan="2">Chunks<br/>total</th>""")
	if ci.metainfomode:
		out.append("""		<th rowspan="2"><a style="cursor:default" title="chunk is in EC format when redundancy level in EC format is higher than number of full copies">EC chunks</a></th>""")
		out.append("""		<th rowspan="2"><a style="cursor:default" title="storage needed to maintain defined redundancy level divided by raw data size">Disks overhead<br/>due to redundancy</a></th>""")
		out.append("""		<th rowspan="2"><a style="cursor:default" title="assumes that full and EC chunks have the same average length and that we save extra copies using redundancy level from EC definition">Space saved<br/>by EC</a></th>""")
	else:
		out.append("""		<th colspan="2">Chunk copies</th>""")
	out.append("""	</tr>""")

	out.append("""	<tr>""")
	out.append("""		<th style="width:70px;">total</th>""")
	out.append("""		<th style="width:70px;">available</th>""")

	out.append("""		<th style="width:140px;">% used</th>""")
	out.append("""		<th style="min-width:50px;">space</th>""")
	out.append("""		<th style="min-width:50px;">files</th>""")
	out.append("""		<th style="min-width:50px;">space</th>""")
	out.append("""		<th style="min-width:50px;">files</th>""")
	out.append("""		<th style="min-width:50px;">all</th>""")
	out.append("""		<th style="min-width:50px;">directories</th>""")
	out.append("""		<th style="min-width:50px;">files</th>""")
	if not ci.metainfomode:
		out.append("""		<th><span data-tt-help="Chunks from all disks: both 'regular' and 'marked for removal'">all</span></th>""")
		out.append("""		<th><span data-tt-help="Only chunks from 'regular' hdd space (excluded 'marked for removal')">regular</span></th>""")
	out.append("""	</tr>""")
	out.append("""	<tr>""")
	out.append("""		<td align="center"><span data-tt-info="%s B">%s</span></td>""" % (decimal_number(ci.totalspace),humanize_number(ci.totalspace,"&nbsp;")))
	out.append("""		<td align="center"><span data-tt-info="%s B">%s</span></td>""" % (decimal_number(ci.availspace),humanize_number(ci.availspace,"&nbsp;")))
	usedpercent=100*(ci.totalspace-ci.availspace)/ci.totalspace if ci.totalspace!=0 else 0
	out.append("""		<td align="center"><div class="PROGBOX"><div class="PROGCOVER" style="width:%.2f%%;"></div><div class="PROGVALUE"><span>%.1f</span></div></div></td>""" % (100.0-usedpercent, usedpercent))
	out.append("""		<td align="center"><a style="cursor:default" title="%s B">%s</a></td>""" % (decimal_number(ci.trspace),humanize_number(ci.trspace,"&nbsp;")))
	out.append("""		<td align="center">%s</td>""" % decimal_number_html(ci.trfiles)) 
	out.append("""		<td align="center"><a style="cursor:default" title="%s B">%s</a></td>""" % (decimal_number(ci.respace),humanize_number(ci.respace,"&nbsp;")))
	out.append("""		<td align="center">%s</td>""" % decimal_number_html(ci.refiles))
	out.append("""		<td align="center">%s</td>""" % decimal_number_html(ci.nodes))
	out.append("""		<td align="center">%s</td>""" % decimal_number_html(ci.dirs))
	out.append("""		<td align="center">%s</td>""" % decimal_number_html(ci.files))
	out.append("""		<td align="center">%s</td>""" % decimal_number_html(ci.chunks))
	if ci.metainfomode:
		if ci.ecchunkspercent!=None:
			out.append("""		<td align="center"><a style="cursor:default" title="full chunks: %u ; EC8 chunks: %u ; EC4 chunks: %u">%.1f&#8239;%%</a></td>""" % (ci.copychunks,ci.ec8chunks,ci.ec4chunks,ci.ecchunkspercent))
		else:
			out.append("""		<td align="center"><a style="cursor:default" title="full chunks: %u ; EC8 chunks: %u ; EC4 chunks: %u">-</a></td>""" % (ci.copychunks,ci.ec8chunks,ci.ec4chunks))
		if ci.dataredundancyratio!=None:
			out.append("""		<td align="center"><a style="cursor:default" title="full copies: %u ; EC8 parts: %u ; EC4 parts: %u ; chunks: %u">+ %.1f&#8239;%%</a></td>""" % (ci.chunkcopies,ci.chunkec8parts,ci.chunkec4parts,(ci.copychunks+ci.ec8chunks+ci.ec4chunks),(ci.dataredundancyratio-1)*100))
		else:
			out.append("""		<td align="center"><a style="cursor:default" title="full copies: %u ; EC8 parts: %u ; EC4 parts: %u ; chunks: %u">-</a></td>""" % (ci.chunkcopies,ci.chunkec8parts,ci.chunkec4parts,(ci.copychunks+ci.ec8chunks+ci.ec4chunks)))
		if ci.savedbyec!=None:
			out.append("""		<td align="center">%s</td>""" % (("about "+humanize_number(ci.savedbyec,"&nbsp;")) if ci.savedbyec>0 else "not yet"))
		else:
			out.append("""		<td align="center">-</td>""")
	else:
		out.append("""		<td align="center">%s</td>""" % decimal_number_html(ci.allcopies))
		out.append("""		<td align="center">%s</td>""" % decimal_number_html(ci.regularcopies))
	out.append("""	</tr>""")
	out.append("""</table>""")
	return out


from common.constants import *
from common.utils import *
from common.utilsgui import *
from common.models import *

def render(dp, fields, vld):
	SCorder = fields.getint("SCorder", 0)
	SCrev = fields.getint("SCrev", 0)
	show_counters = fields.getint("SCcounters", 0) # show inodes and chunk counters columns?

	usingEC = dp.master().version_at_least(4,5,0)
	show_arch_min_size = dp.master().version_at_least(4,34,0)
	show_labelmode_overrides = dp.master().has_feature(FEATURE_LABELMODE_OVERRIDES) # v.4.53.0
	show_export_group_and_priority = dp.master().has_feature(FEATURE_SCLASSGROUPS) # v.4.57.0
	out = []
	out.append("""<div class="tab_title">Storage classes""")
	out.append("""</div>""")
	out.append("""<table class="acid_tab acid_tab_zebra_C1_C2 acid_tab_storageid_mfssc" id="mfssc">""")
	if usingEC:
		out.append("""	<tr>""")
		colspan = 6 if show_export_group_and_priority else 4
		out.append("""		<th rowspan="2" colspan="%u" class="knob-cell">""" % (colspan,))
		options=[(-57,"Hide","counters","document.location.href='%s&SCcounters=0'" % fields.createrawlink({"SCcounters":""})),
				(57, "Show","counters","document.location.href='%s&SCcounters=1'" % fields.createrawlink({"SCcounters":""}))]
		selected = 1 if show_counters else 0
		out.append(html_knob_selector("sccnt_vis",9,(150,28),(80,14),options,True,selected))
		out.append("""		</th>""")
		if show_counters:
			out.append("""		<th colspan="2" rowspan="2" class="acid_tab_skip sccnt_vis1">Inodes<br/>count</th>""")
		out.append("""		<th colspan="%d" rowspan="1" class="acid_tab_skip sccnt_vis1">States transitions</th>""" % (3 + show_arch_min_size))
		out.append("""		<th rowspan="4">Default<br/>class<br/>match<br/>mode</th>""")
		if show_counters:
			out.append("""		<th rowspan="1" colspan="%u" class="acid_tab_skip">States definitions and counters</th>""" % (12 + show_labelmode_overrides))
		else:
			out.append("""		<th rowspan="1" colspan="%u" class="acid_tab_skip">States definitions</th>""" % (6 + show_labelmode_overrides))
		out.append("""	</tr>""")

		out.append("""	<tr>""")
		out.append("""		<th colspan="%d" class="acid_tab_skip">KEEP &harr; ARCHIVE</th>""" % (2 + show_arch_min_size))
		out.append("""		<th class="acid_tab_skip">&rarr; TRASH</th>""")
		if show_labelmode_overrides:
			out.append("""		<th rowspan="3" class="acid_tab_skip">State<br/>match<br/>mode</th>""")
		out.append("""		<th rowspan="3" class="acid_tab_skip">State</th>""")
		out.append("""		<th rowspan="2" colspan="3" class="acid_tab_skip">Redundancy</th>""")
		out.append("""		<th rowspan="2" colspan="2" class="acid_tab_skip">Storage</th>""")
		if show_counters:
			out.append("""		<th rowspan="1" colspan="6" class="acid_tab_skip">Chunks count</th>""")
		out.append("""	</tr>""")

		out.append("""	<tr>""")
		out.append("""		<th rowspan="2" class="acid_tab_enumerate">#</th>""")
		out.append("""		<th rowspan="2">Id</th>""")
		out.append("""		<th rowspan="2" class="acid_tab_alpha">Name</th>""")
		if show_export_group_and_priority:
			out.append("""		<th rowspan="2">Join<br/>priority</th>""")
			out.append("""		<th rowspan="2">Export<br/>group</th>""")
		out.append("""		<th rowspan="2">Admin<br/>only</th>""")
		if show_counters:
			out.append("""		<th rowspan="2">files</th>""")
			out.append("""		<th rowspan="2">dirs</th>""")
		out.append("""		<th rowspan="2">mode</th>""")
		out.append("""		<th rowspan="2" style="min-width:32px;">delay</th>""")
		if show_arch_min_size:
			out.append("""		<th rowspan="2" style="min-width:32px;">min<br/>size</th>""")
		out.append("""		<th rowspan="2">min trash<br/>retention</th>""") #min trash retension
		if show_counters:
			out.append("""		<th colspan="2" class="acid_tab_skip">deficient</th>""")
			out.append("""		<th colspan="2" class="acid_tab_skip">stable</th>""")
			out.append("""		<th colspan="2" class="acid_tab_skip">over goal</th>""")
		out.append("""	</tr>""")

		out.append("""	<tr>""")
		out.append("""		<th class="acid_tab_skip">level</th>""")
		out.append("""		<th class="acid_tab_skip">format</th>""")
		out.append("""		<th class="acid_tab_skip">achievable</th>""")
		out.append("""		<th class="acid_tab_skip">labels expressions</th>""")
		out.append("""		<th class="acid_tab_skip">distribution</th>""")
		if show_counters:
			out.append("""		<th class="acid_tab_skip" style="min-width:32px;">copy</th>""")
			out.append("""		<th class="acid_tab_skip" style="min-width:32px;">EC</th>""")
			out.append("""		<th class="acid_tab_skip" style="min-width:32px;">copy</th>""")
			out.append("""		<th class="acid_tab_skip" style="min-width:32px;">EC</th>""")
			out.append("""		<th class="acid_tab_skip" style="min-width:32px;">copy</th>""")
			out.append("""		<th class="acid_tab_skip" style="min-width:32px;">EC</th>""")
		out.append("""	</tr>""")
	else:
		show_counters = 1 # always show counters for older versions
		out.append("""	<tr>""")
		out.append("""		<th rowspan="3" class="acid_tab_enumerate">#</th>""")
		out.append("""		<th rowspan="2">Id</th>""")
		out.append("""		<th rowspan="2" classname="acid_tab_alpha">Name</th>""")
		out.append("""		<th rowspan="2">Admin only</th>""")
		out.append("""		<th colspan="2" class="acid_tab_skip">Inodes count</th>""")
		out.append("""		<th rowspan="2">Archive mode</th>""")
		out.append("""		<th rowspan="2">Archive delay</th>""")
		out.append("""		<th rowspan="2">Min trashretention</th>""")
		out.append("""		<th rowspan="2">Labels mode</th>""")
		out.append("""		<th colspan="5" class="acid_tab_skip">State definitions</th>""")
		out.append("""		<th colspan="5" class="acid_tab_skip">Chunks count</th>""")
		out.append("""	</tr>""")
		out.append("""	<tr>""")
		out.append("""		<th>files</th>""")
		out.append("""		<th>dirs</th>""")
		out.append("""		<th class="acid_tab_skip">state</th>""")
		out.append("""		<th class="acid_tab_skip">format</th>""")
		out.append("""		<th class="acid_tab_skip">achievable</th>""")
		out.append("""		<th class="acid_tab_skip">labels expressions</th>""")
		out.append("""		<th class="acid_tab_skip">distribution</th>""")
		out.append("""		<th class="acid_tab_skip">deficient</th>""")
		out.append("""		<th class="acid_tab_skip">stable</th>""")
		out.append("""		<th class="acid_tab_skip">over goal</th>""")
		out.append("""	</tr>""")
	
	for sc in dp.get_sclasses(SCorder, SCrev):
		out.append("""	<tr>""")
		rowcnt = len(sc.states)
		out.append("""		<td rowspan="%u" align="right"></td>""" % (rowcnt,))
		out.append("""		<td rowspan="%u" align="right">%u</td>""" % (rowcnt,sc.sclassid))
		# name
		sclassname=htmlentities(sc.sclassname)
		if sc.sclassdesc!=None and sc.sclassdesc!="":
			desc = htmlentities(sc.sclassdesc)
			out.append("""		<td rowspan="%u" class="word-wrap" style="min-width:5em;max-width:15em;white-space:normal;"><span class="center bold">%s</span><br/><span class="em9" style="text-overflow: ellipsis;">%s</span></td>""" % (rowcnt,sclassname,desc))
		else:
			out.append("""		<td rowspan="%u" class="center bold">%s</td>""" % (rowcnt,sclassname))
		# priority and export group
		if show_export_group_and_priority:
			out.append("""		<td rowspan="%u" align="center">%u</td>""" % (rowcnt,sc.priority))
			out.append("""		<td rowspan="%u" align="center">%u</td>""" % (rowcnt,sc.export_group))
		# admin only
		out.append("""		<td rowspan="%u" align="center">%s</td>""" % (rowcnt,sc.get_admin_only_str().capitalize()))
		# counters
		if show_counters:
			out.append("""		<td rowspan="%u" align="right">%s</td>""" % (rowcnt,decimal_number_html(sc.files)))
			out.append("""		<td rowspan="%u" align="right">%s</td>""" % (rowcnt,decimal_number_html(sc.dirs)))
		# mode
		out.append("""		<td rowspan="%u" align="center"><span class="sortkey">%u </span>%s</td>""" % (rowcnt,sc.arch_mode,mode2html(sc)))
		# delay		
		info = "sc_arch_delay0" if sc.arch_delay==0 else ""
		out.append("""		<td rowspan="%u" align="center"><span class="sortkey">%u </span><span data-tt="%s">%s</span></td>""" % (rowcnt,sc.arch_delay,info,archdelay2html(sc)))
		# min size
		if show_arch_min_size:
			info = ("%s B" % decimal_number(sc.arch_min_size)) if sc.arch_min_size>0 else "sc_arch_min_size0"
			out.append("""		<td rowspan="%u" align="center"><span class="sortkey">%u </span><span data-tt="%s">%s</span>""" % (rowcnt,sc.arch_min_size,info,humanize_number(sc.arch_min_size,nnbsp)))
		# min trash retention
		issues = vld.check_sc_mintrashretention(sc)
		out.append("""		<td rowspan="%u" align="center"><span class="sortkey">%u </span>%s</td>""" % (rowcnt,sc.min_trashretention,issues.span_noicon(mintrashretention2html(sc))))
		# labels match mode
		overridden = False
		clrcls = ""
		border = ""
		if show_labelmode_overrides:
			for st in sc.states:
				if st.defined and st.labelsmodeover>=0 and st.labelsmodeover<=2:
					overridden = True
					break
			if overridden: clrcls = "GRAY"
			border = """ style="border-right: none;" """
		out.append("""		<td rowspan="%u" align="center" %s><span class="%s">%s</span></td>""" % (rowcnt,border,clrcls,scfitmode2html(sc)))
		
		# states: CREATE, KEEP, ARCHIVE, TRASH
		newrow = 0
		group_states = sc.isKeepOnly() # group states if only KEEP is defined (this is often the case)
		for line,st in enumerate(sc.states):
			show_line = not group_states or (group_states and line==0)
			rowspan = 4 if group_states else 1
			if newrow:
				out.append("""	</tr>""")
				out.append("""	<tr>""")
			else:
				newrow=1
			colorclass = "" if st.defined else "GRAY"
			wcolorclass = "WARNING" if st.defined or group_states else "GRAY"
			ecolorclass = "ERROR" if st.defined or group_states else "GRAY"
			if show_line:
				if show_labelmode_overrides:
					clrcls = "" if overridden else "GRAY"
					arrow = "&rarr;" if (not (st.labelsmodeover>=0 and st.labelsmodeover<=2 and st.defined)) else "&nbsp;"
					out.append("""		<td rowspan="%u" align="left" style="border-left: none;"><span style="margin-left:-5px;" class="GRAY monospace">%s</span>&#8239;<span class="%s">%s</span></td>""" % (rowspan,arrow,clrcls,stfitmode2html(st,sc)))
			# state name
			out.append("""		<td align="center"><span class="%s">%s</span></td>""" % (colorclass,st.fullname))

			# level 
			if show_line:
				colorclass = "" if group_states else colorclass
				if usingEC and st.ec_level!=None and st.ec_level>0:
					out.append("""		<td rowspan="%u" align="center"><span class="%s bold">%s</span></td>""" % (rowspan,colorclass,st.ec_chksum_parts))
				else:
					copies = len(st.labellist)
					if usingEC:
						out.append("""		<td rowspan="%u" align="center"><span class="%s bold">%s</span></td>""" % (rowspan,colorclass,copies-1))
				# format 
				out.append("""		<td rowspan="%u" align="left"><span class="%s">&nbsp;%s</span></td>""" % (rowspan,colorclass,format2html(usingEC,st)))

				# achievable
				if st.canbefulfilled==3:
					out.append("""		<td rowspan="%u" align="center"><span class="%s">Yes</span></td>""" % (rowspan,colorclass))
				elif st.canbefulfilled==2:
					out.append("""		<td rowspan="%u" align="center"><span class="%s" data-tt="sc_achievable_overloaded">Overloaded</span></td>""" % (rowspan,wcolorclass))
				elif st.canbefulfilled==1:
					out.append("""		<td rowspan="%u" align="center"><span class="%s" data-tt="sc_achievable_nospace">No space</span></td>""" % (rowspan,wcolorclass))
				elif st.canbefulfilled==4:
					out.append("""		<td rowspan="%u" align="center"><span class="%s" data-tt="sc_achievable_keeponly">EC on hold</span></td>""" % (rowspan,wcolorclass))
				else:
					out.append("""		<td rowspan="%u" align="center"><span class="%s" data-tt="sc_achievable_no">No</span></td>""" % (rowspan,ecolorclass))
				
				# labels
				(tooltip, txt) = labels2html(usingEC,sc,st)
				out.append("""		<td rowspan="%u" align="center"><span data-tt-info="%s">%s</span></td>""" % (rowspan, tooltip, txt))

				# distribution
				distribution = uniqmask_to_str(st.uniqmask)
				distribution = "random" if distribution=='-' else distribution
				out.append("""		<td rowspan="%u" align="center"><span class="%s">%s</span></td>""" % (rowspan,colorclass,distribution))

			# chunk counters
			if show_counters:
				if usingEC:
					if st.counters[0]!=None and st.counters[1]!=None and st.counters[2]!=None and st.counters[3]!=None and st.counters[4]!=None and st.counters[5]!=None and st.defined:
						if st.ec_level!=None and st.ec_level>0:
							# EC state
							out.append("""		<td align="right"><span class="WARNING" data-tt="sc_deficient_chunks_ec_as_copy">%s</span></td>""" % (("%s" % decimal_number_html(st.counters[0])) if st.counters[0]>0 else "&nbsp;"))
							out.append("""		<td align="right"><span class="WARNING" data-tt="sc_deficient_chunks_same_format">%s</span></td>""" % (("%s" % decimal_number_html(st.counters[1])) if st.counters[1]>0 else "&nbsp;"))
							out.append("""		<td align="right"><span class="WRONGFORMAT" data-tt="sc_ec_as_copy">%s</span></td>""" % (("%s" % decimal_number_html(st.counters[2])) if st.counters[2]>0 else "&nbsp;"))
							out.append("""		<td align="right"><span class="OK">%s</span></td>""" % decimal_number_html(st.counters[3]))
						else:
							# copies state
							out.append("""		<td align="right"><span class="WARNING" data-tt="sc_deficient_chunks_same_format">%s</span></td>""" % (("%s" % decimal_number_html(st.counters[0])) if st.counters[0]>0 else "&nbsp;"))
							out.append("""		<td align="right"><span class="WARNING" data-tt="sc_deficient_chunks_copy_as_ec">%s</span></td>""" % (("%s" % decimal_number_html(st.counters[1])) if st.counters[1]>0 else "&nbsp;"))
							out.append("""		<td align="right"><span class="OK">%s</span></td>""" % decimal_number_html(st.counters[2]))
							out.append("""		<td align="right"><span class="WRONGFORMAT" data-tt="sc_copy_as_ec">%s</span></td>""" % (("%s" % decimal_number_html(st.counters[3])) if st.counters[3]>0 else "&nbsp;"))
						out.append("""		<td align="right"><span class="NOTICE">%s</span></td>""" % (("%s" % decimal_number_html(st.counters[4])) if st.counters[4]>0 else "&nbsp;"))
						out.append("""		<td align="right"><span class="NOTICE">%s</span></td>""" % (("%s" % decimal_number_html(st.counters[5])) if st.counters[5]>0 else "&nbsp;"))
					else:
						out.append("""		<td align="center"></td>""")
						out.append("""		<td align="center"></td>""")
						out.append("""		<td align="center"></td>""")
						out.append("""		<td align="center"></td>""")
						out.append("""		<td align="center"></td>""")
						out.append("""		<td align="center"></td>""")
				else:
					if st.counters[0]!=None and st.counters[1]!=None and st.counters[2]!=None and st.defined:
						out.append("""		<td align="right"><span class="WARNING">%s</span></td>""" % (("%s" % decimal_number_html(st.counters[0])) if st.counters[0]>0 else "&nbsp;"))
						out.append("""		<td align="right"><span class="OK">%s</span></td>""" % decimal_number_html(st.counters[1]))
						out.append("""		<td align="right"><span class="NOTICE">%s</span></td>""" % (("%s" % decimal_number_html(st.counters[2])) if st.counters[2]>0 else "&nbsp;"))
					else:
						out.append("""		<td align="center"></td>""")
						out.append("""		<td align="center"></td>""")
						out.append("""		<td align="center"></td>""")

		out.append("""	</tr>""")
	out.append("""</table>""")

	return out

def format2html(usingEC,st):
	if usingEC and st.ec_level!=None and st.ec_level>0:
		return "EC&#8239;%u&#8239;+%u</span></td>" % (st.ec_data_parts,st.ec_chksum_parts)
	else:
		copies = len(st.labellist)
		copies_txt = "copies" if copies>1 else "copy"
		return "%u %s" % (copies,copies_txt)


# get mode cell content
def mode2html(sc):
	arch_mode_list = []
	if sc.arch_mode&SCLASS_ARCH_MODE_CHUNK:
		mode='Per chunk<br/><span style="display: block; margin-top: -4px;">&rarr;</span>'
		arch_mode_list.append("chunk's<br/>mtime")
	elif sc.arch_mode&SCLASS_ARCH_MODE_FAST:
		mode='Fast<br/><span style="display: block; margin-top: -4px;">&rarr;</span>'
	else:
		if sc.arch_mode&SCLASS_ARCH_MODE_REVERSIBLE:
			mode='Reversible<br/><span style="display: block; margin-top: -4px;">&harr;</span>'
		else:
			mode='Oneway<br/><span style="display: block; margin: -4px 0 4px 0;">&rarr;</span>'
		align = "&nbsp;" if sc.arch_mode&SCLASS_ARCH_MODE_MTIME else ""
		if sc.arch_mode&SCLASS_ARCH_MODE_CTIME:
			arch_mode_list.append(align+"ctime")
		if sc.arch_mode&SCLASS_ARCH_MODE_MTIME:
			arch_mode_list.append("mtime")
		if sc.arch_mode&SCLASS_ARCH_MODE_ATIME:
			arch_mode_list.append(align+"atime")
	arch_mode_str = "<br/>".join(arch_mode_list)

	if len(arch_mode_str)>0:
		return mode+arch_mode_str
	else:
		return mode

def scfitmode2html(sc):
	return "Loose" if sc.labels_mode==0 else "Standard" if sc.labels_mode==1 else "Strict"

def stfitmode2html(st,sc):
	return "Loose" if st.labelsmodeover==0 else "Standard" if st.labelsmodeover==1 else "Strict" if st.labelsmodeover==2 else scfitmode2html(sc)


# get delay cell content
def archdelay2html(sc):
	if (sc.arch_delay>0 and (sc.arch_mode&SCLASS_ARCH_MODE_FAST)==0):
			return hours2html(sc.arch_delay)
	if sc.arch_mode&SCLASS_ARCH_MODE_FAST!=0:
		return ""
	return "0"

def mintrashretention2html(sc):
	if not sc.defined_trash and sc.min_trashretention==0:
		return 'n/a'
	if sc.min_trashretention>0:
		return hours2html(sc.min_trashretention)
	return "0"

# get labels cell class, tooltip and content
def labels2html(usingEC, sc, st):
	def label2str(label, literals=['any server', ' and ', ' or ', ' not']):
		if sc.labels_ver==4:
			return labelexpr_to_str(label, literals)
		else:
			return labelmasks_to_str(label)
	
	if sc.isKeepOnly():
		title = "Labels expressions (all states)"
	else:
		title = "%s labels expressions" % st.fullname
	tooltip = "<table class='labels-table'>"
	lead = "Class " if len(sc.sclassname)<5 else ""
	tooltip += "<tr><th colspan='3'>%s<span class='NOTICE em11'>%s</span></th></tr>" % (lead,sc.sclassname)
	tooltip += "<tr><th colspan='3'>%s<hr></th></tr>" % title
	tooltip += "<tr><th class='em11 NOTICE'>%s</th><th>Labels<br/>expression</th><th>Available<br/>servers</th></tr>" % format2html(usingEC,st)
	labelsarr = []
	i = 1
	odd = True
	not_enough_servers = False
	only_one_server = False
	if st.isEC():
		for i in range(2):
			if len(st.labellist)>i:
				_,matchingservers,labelexpr = st.labellist[i]
			elif len(st.labellist)>0:
				_,matchingservers,labelexpr = st.labellist[0]
			else:
				matchingservers = 0
				labelexpr = []
			if matchingservers==0:
				not_enough_servers = True
			srvavailclr = ""
			tooltip += "<tr class='%s'>" % ("odd" if i==0 else "even")
			if i==0:
				tooltip += "<td class='left' style='padding-left:10px; min-width:90px;'>%s data parts</td>" % str(st.ec_data_parts)
			else:
				tooltip += "<td class='left' style='padding-left:10px; min-width:90px;'>%s parity part%s</td>" % (str(st.ec_chksum_parts), "s" if st.ec_chksum_parts>1 else "")
			tooltip += "<td>%s</td>" % htmlentities(label2str(labelexpr))
			srvmatchclr = "ERROR" if matchingservers==0 else "WARNING" if matchingservers==1 else srvavailclr
			tooltip += "<td class='right' style='padding-right:20px;'><span class='%s'>%s</span> <span class='%s'>of %s</span></td>" % (srvmatchclr, str(matchingservers), srvavailclr, str(sc.availableservers))
			tooltip += "</tr>"

	else:
		for _,matchingservers,labelexpr in st.labellist:
			if matchingservers==0:
				not_enough_servers = True
			elif matchingservers==1:
				only_one_server = True
			if odd:
				tooltip += "<tr class='odd'>"
			else:
				tooltip += "<tr class='even'>"
			srvavailclr = "ERROR" if i>sc.availableservers else ""
			tooltip += "<td class='right %s' style='padding-right:10px;'>%s<span class='em8'>%s</span> copy</td>" % (srvavailclr, str(i), ordinal_suffix(i))
			tooltip += "<td>%s</td>" % htmlentities(label2str(labelexpr))
			srvmatchclr = "ERROR" if matchingservers==0 else "WARNING" if matchingservers==1 else srvavailclr
			tooltip += "<td class='right' style='padding-right:20px;'><span class='%s'>%s</span> <span class='%s'>of %s</span></td>" % (srvmatchclr, str(matchingservers), srvavailclr, str(sc.availableservers))
			tooltip += "</tr>"
			i+=1
			odd = not odd
	tooltip += "<tr><td colspan='3'>"
	if st.isEC() and st.ec_data_parts+st.ec_chksum_parts>sc.availableservers:
		tooltip += "<hr><b>Warning</b>: There are more data and parity parts (%s+%s=%s) than available servers (%s)." % (st.ec_data_parts, st.ec_chksum_parts, st.ec_data_parts+st.ec_chksum_parts, sc.availableservers)
		not_enough_servers = False # do not show this warning any more
	if not st.isEC() and len(st.labellist)>sc.availableservers:
		tooltip += "<hr><b>Warning:</b> There are more expected copies (labels expressions) than available servers."
	if not_enough_servers:
		tooltip += "<hr><b>Warning:</b> There are not enough servers to fulfill the labels expression for this state."
	elif only_one_server:
		tooltip += "<hr><b>Warning:</b> There is only single server that matches some labels expressions for this state."
	if sc.isKeepOnly():
		tooltip += "<hr>This storage class has only definition for the KEEP state. All other states (CREATE, ARCHIVE and TRASH) inherit the labels expression from the KEEP state."
	elif not st.defined:
		tooltip += "<hr>This state (%s) is not explicitly defined in the storage class, it inherits the labels expression from the KEEP state." % st.fullname
	tooltip += "</td></tr>"
	tooltip += "</table>"

	i = 1
	foldedlist = labellist_fold(st.labellist)
	for _,matchingservers,label in foldedlist:
		
		if st.defined or sc.isKeepOnly():
			cls_expr = ""
			if int(matchingservers)==0:
				clz = "ERROR"
			elif int(matchingservers)==1:
				clz = "WARNING"
			else:
				clz = "GRAY"
		else:
			clz = "GRAY"	
			cls_expr = "GRAY"
		labelstr = label2str(label, literals=['*', '', '|', '~'])
		if labelstr=="*":
			labelstr = "Any server"
		else:
			labelstr = """<span class="monospace">%s</span>""" % htmlentities(labelstr)
		if (len(foldedlist)>1):
			if st.isEC():
				txt = "data:" if i==1 else "parity:"
			else:
				txt = "%s%s:" % (i,ordinal_suffix(i))
		else:
			txt = "all:"
		labelstr = """<span class="bold em9 %s">%s&#8239;</span><span class="%s">%s</span>""" % (clz, txt, cls_expr, labelstr)
		labelsarr.append(labelstr)
		i+=1
	
	txt = "&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;".join(labelsarr)
	return (tooltip, txt)



class Organization:
	def __init__(self, dp, guimode=False, ajax_request=False):
		self.sectionset = None
		self.subsectionset = None
		self.guimode = guimode
		cl = dp.cluster
		# Define menu & subsections structure
		if guimode:
			if cl.leaderfound():
				self.menu_tree = [
					["ST", ["ST"]],
					["IN", ["IG", "IM", "LI", "IC", "FL", "CL", "MU"]],
					["CS", ["CS", "MB"]],
					["HD", ["HD"]],
					["EX", ["EX"]],
					["MD", ["MS", "MO", "OF"]],
					["RP", ["SC", "PA"]],
					["QU", ["QU"]],
					["XC", ["MC", "CC"]]
				]
			elif cl.electfound():
				self.menu_tree = [
					["ST", ["ST"]],
					["IN", ["IG", "IM", "LI", "IC", "FL", "CL", "MU"]],
					["CS", ["CS", "MB"]],
					["HD", ["HD"]],
					["EX", ["EX"]],
					["QU", ["QU"]],
					["XC", ["MC", "CC"]]
				]
			elif cl.followerfound() or cl.usurperfound():
				self.menu_tree = [
					["IN", ["IG", "IM", "LI", "IC", "FL", "CL", "MU"]],
					["MC", ["MC"]]
				]
			else:
				self.menu_tree = [
					["IN", ["IG", "IM", "LI", "IC", "FL", "CL", "MU"]]
				]
		else:
			# Define menu & subsections structure
			if cl.leaderfound():
				self.menu_tree = [
					["IN", ["IG", "IM", "LI", "IC", "IL", "MF", "MU"]],
					["CS", ["CS"]],
					["MB", ["MB"]],
					["HD", ["HD"]],
					["EX", ["EX"]],
					["MD", ["MS", "OF", "AL", "MO"]],
					["RP", ["SC", "PA"]],
					["RS", ["SC", "PA", "OF", "AL"]], # backwards compatibility
					["QU", ["QU"]],
					["MC", ["MC"]],
					["CC", ["CC"]]
				]
			elif cl.electfound():
				self.menu_tree = [
					["IN", ["IG", "IM", "LI", "IC", "IL", "MF", "MU"]],
					["CS", ["CS"]],
					["MB", ["MB"]],
					["HD", ["HD"]],
					["EX", ["EX"]],
					["QU", ["QU"]],
					["MC", ["MC"]],
					["CC", ["CC"]]
				]
			elif cl.followerfound() or cl.usurperfound():
				self.menu_tree = [
					["IN", ["IG", "IM", "LI", "IC", "IL", "MF", "MU"]],
					["MC", ["MC"]]
				]
			elif cl.anyfound():
				self.menu_tree = [
					["IN", ["IG", "IM", "LI", "IC", "IL", "MF", "MU"]]
				]
			else:
				self.menu_tree = None


		# sections names and render conditions
		self.sections_defs = {
			"ST":("Status",										 lambda: guimode and (cl.leaderfound() or cl.electfound())),
			"IN":("Info",                      lambda: True),
				"IG":("Cluster Summary",           lambda: cl.master()!=None),
				"IM":("Metadata Servers (Masters)",lambda: len(dp.get_masterservers(None))>0),
				"MB":("Metadata backup loggers",   lambda: cl.leaderfound() and cl.master()!=None),
				"LI":("Licence Info",              lambda: cl.leaderfound() and cl.master()!=None and cl.master().is_pro()),
				"IC":("Chunk Matrix Table",        lambda: cl.leaderfound()),
				"IL":("Self-check Loops",          lambda: cl.leaderfound() and not guimode),
				"MF":("Missing Files",             lambda: cl.leaderfound() and ((cl.master().version_at_least(2,0,66) and cl.master().version_less_than(3,0,0)) or cl.master().version_at_least(3,0,19))),
				"FL":("Filesystem Self-check Loop",lambda: cl.leaderfound() and guimode), #GUI only
				"CL":("Chunks Housekeeping Loop",  lambda: cl.leaderfound() and guimode), #GUI only
				"MU":("Memory Usage",              lambda: cl.leaderfound() and cl.master()!=None),
			"CS":("Chunkservers",                lambda: cl.master()!=None),
			"HD":("Disks",                     lambda: cl.master()!=None),
			"EX":("Exports",                   lambda: cl.master()!=None),
			"MD":("Mounts",                    lambda: cl.leaderfound()),
				"MS":("Parameters",                lambda: cl.leaderfound()),
				"MO":("Operations",                lambda: cl.leaderfound()),
				"OF":("Open files and locks",      lambda: cl.leaderfound()),
				"AL":("Acquired locks",            lambda: cl.leaderfound() and not guimode), #CLI only, for GUI "AL" is in "OF"
			"RP":("Redundancy",                lambda: cl.leaderfound()),
				"SC":("Storage Classes",           lambda: cl.leaderfound()),
				"PA":("Override Patterns",         lambda: cl.leaderfound() and cl.master().version_at_least(4,2,0)),
			"RS":("Resources",                 lambda: cl.leaderfound()),# backwards compatibility
			"QU":("Quotas",                    lambda: cl.master()!=None),
			"XC":("Charts",                    lambda: cl.master()!=None and not ajax_request),
				"MC":("Master servers",            lambda: cl.master()!=None and not ajax_request),
				"CC":("Chunkservers",              lambda: cl.master()!=None and not ajax_request)
		}

	def set_sections(self, sectionset, subsectionset=None):
		self.sectionset = sectionset
		self.subsectionset = subsectionset

	def get_default_section(self):
		return self.menu_tree[0][0]
	
	# returns list of subsections ids for given menu section
	def menu_subitems(self, section):
		for menu_item in self.menu_tree:
			if menu_item[0]==section:
				return menu_item[1]
		return []

	# Decide if section should be rendered
	def shall_render(self, section):
		if self.sectionset==None:
			raise Exception("Organization object is not initialized properly: set_sections() must be called before shall_render()")
		
		# render section if it's in the sectionset
		effective_sectionset = set(self.sectionset)

		if self.guimode:
			if self.subsectionset==None:
				raise Exception("Organization object is not initialized properly: set_sections() must be called before shall_render()")
			# adjust default subsections
			if "RP" in self.sectionset: 
				if len(self.subsectionset.intersection(set(self.menu_subitems("RP"))))==0:
					self.subsectionset |= set(("SC",)) # add Storage Classes section by default if Redundancy Policy is selected
			if "IN" in self.sectionset:
				if len(self.subsectionset.intersection(set(self.menu_subitems("IN"))))==0:
					self.subsectionset |= set((self.menu_subitems("IN"))) # add all subsections by default if Info is selected
			# render subsection only if 
			#    its main section is visible and it's in the subsectionset 
			# or its main section is visible and there no subsections for this main section specified
			for main_section in self.sectionset:
				subitemset = set(self.menu_subitems(main_section))
				itsmainsectionvisible = section in subitemset
				if itsmainsectionvisible and section in self.subsectionset:
					effective_sectionset |= set((section,))
					break
		else:
			# for CLI render all subsections if the main section is selected
			for s in self.sectionset:
				effective_sectionset |= set(self.menu_subitems(s))

		if section in effective_sectionset:
			return self.sections_defs[section][1]() # render this section if condition is met

		return False # don't render






from common.utils import *
from common.utilsgui import *

def render(dp, fields, vld):
	PAorder = fields.getint("PAorder", 0)
	PArev = fields.getint("PArev", 0)
	
	out = []
	out.append("""<div class="tab_title">Override patterns</div>""")
	out.append("""<table class="acid_tab acid_tab_zebra_C1_C2 acid_tab_storageid_mfspatterns">""")
	out.append("""	<tr>""")
	out.append("""		<th class="acid_tab_enumerate">#</th>""")
	out.append("""		<th>Pattern (GLOB)</th>""")
	out.append("""		<th>euid</th>""")
	out.append("""		<th>egid</th>""")
	out.append("""		<th>Priority</th>""")
	out.append("""		<th>Storage&nbsp;class</th>""")
	out.append("""		<th>Trash&nbsp;retention</th>""")
	out.append("""		<th>Extra&nbsp;attributes</th>""")
	out.append("""	</tr>""")

	for op in dp.get_opatterns(PAorder, PArev):
		if op.sclassname==None:
			op.sclassname = "-"
		out.append("""	<tr>""")
		out.append("""		<td align="right"></td>""")
		out.append("""		<td align="center">%s</td>""" % htmlentities(op.globname))
		out.append("""		<td align="center">%s</td>""" % op.euidstr)
		out.append("""		<td align="center">%s</td>""" % op.egidstr)
		out.append("""		<td align="center">%u</td>""" % op.priority)
		out.append("""		<td align="center">%s</td>""" % htmlentities(op.sclassname))
		if op.trashretention==None:
			out.append("""		<td align="center"><span class="sortkey">0 </span>-</td>""")
		else:
			out.append("""		<td align="center"><span class="sortkey">%u </span><a style="cursor:default" title="%s">%s</a></td>""" % (op.trashretention,hours_to_str(op.trashretention),timeduration_to_shortstr(op.trashretention*3600)))
		out.append("""		<td align="center"><span class="sortkey">%u </span>%s</td>""" % (op.seteattr|op.clreattr,op.eattrstr))
		out.append("""	</tr>""")

	out.append("""</table>""")
	return out


import sys
import traceback


from common.constants import *
from common.utils import *
from common.utilsgui import *
from common.models import *
from common.conn import MFSConn

def process_commands(cl,fields,html_title):
	# commands in CGI mode
	cmd_success = -1
	if "CSremove" in fields:
		cmd_success = 0
		tracedata = ""
		if cl.leaderfound():
			try:
				serverdata = fields.getvalue("CSremove").split(":")
				if len(serverdata)==2:
					csip = list(map(int,serverdata[0].split(".")))
					csport = int(serverdata[1])
					if len(csip)==4:
						data,length = cl.master().command(CLTOMA_CSSERV_COMMAND,MATOCL_CSSERV_COMMAND,struct.pack(">BBBBBH",MFS_CSSERV_COMMAND_REMOVE,csip[0],csip[1],csip[2],csip[3],csport))
						if length==1:
							status = (struct.unpack(">B",data))[0]
							cmd_success = 1
			except Exception:
				tracedata = traceback.format_exc()
		url = fields.createrawlink({"CSremove":""})
	elif "CSbacktowork" in fields:
		cmd_success = 0
		tracedata = ""
		if cl.leaderfound() and cl.master().version_at_least(1,6,28):
			try:
				serverdata = fields.getvalue("CSbacktowork").split(":")
				if len(serverdata)==2:
					csip = list(map(int,serverdata[0].split(".")))
					csport = int(serverdata[1])
					if len(csip)==4:
						data,length = cl.master().command(CLTOMA_CSSERV_COMMAND,MATOCL_CSSERV_COMMAND,struct.pack(">BBBBBH",MFS_CSSERV_COMMAND_BACKTOWORK,csip[0],csip[1],csip[2],csip[3],csport))
						if length==1:
							status = (struct.unpack(">B",data))[0]
							cmd_success = 1
			except Exception:
				tracedata = traceback.format_exc()
		url = fields.createrawlink({"CSbacktowork":""})
	elif "CSmaintenanceon" in fields:
		cmd_success = 0
		tracedata = ""
		if cl.leaderfound() and cl.master().version_at_least(2,0,11):
			try:
				serverdata = fields.getvalue("CSmaintenanceon").split(":")
				if len(serverdata)==2:
					csip = list(map(int,serverdata[0].split(".")))
					csport = int(serverdata[1])
					if len(csip)==4:
						data,length = cl.master().command(CLTOMA_CSSERV_COMMAND,MATOCL_CSSERV_COMMAND,struct.pack(">BBBBBH",MFS_CSSERV_COMMAND_MAINTENANCEON,csip[0],csip[1],csip[2],csip[3],csport))
						if length==1:
							status = (struct.unpack(">B",data))[0]
							cmd_success = 1
			except Exception:
				tracedata = traceback.format_exc()
		url = fields.createrawlink({"CSmaintenanceon":""})
	elif "CSmaintenanceoff" in fields:
		cmd_success = 0
		tracedata = ""
		if cl.leaderfound() and cl.master().version_at_least(2,0,11):
			try:
				serverdata = fields.getvalue("CSmaintenanceoff").split(":")
				if len(serverdata)==2:
					csip = list(map(int,serverdata[0].split(".")))
					csport = int(serverdata[1])
					if len(csip)==4:
						data,length = cl.master().command(CLTOMA_CSSERV_COMMAND,MATOCL_CSSERV_COMMAND,struct.pack(">BBBBBH",MFS_CSSERV_COMMAND_MAINTENANCEOFF,csip[0],csip[1],csip[2],csip[3],csport))
						if length==1:
							status = (struct.unpack(">B",data))[0]
							cmd_success = 1
			except Exception:
				tracedata = traceback.format_exc()
		url = fields.createrawlink({"CSmaintenanceoff":""})
	elif "MSremove" in fields:
		cmd_success = 0
		tracedata = ""
		if cl.leaderfound():
			try:
				sessionid = int(fields.getvalue("MSremove"))
				data,length = cl.master().command(CLTOMA_SESSION_COMMAND,MATOCL_SESSION_COMMAND,struct.pack(">BL",MFS_SESSION_COMMAND_REMOVE,sessionid))
				if length==1:
					status = (struct.unpack(">B",data))[0]
					cmd_success = 1
			except Exception:
				tracedata = traceback.format_exc()
		url = fields.createrawlink({"MSremove":""})
	elif "CSclearerrors" in fields:
		cmd_success = 0
		tracedata = ""
		try:
			serverdata = fields.getvalue("CSclearerrors").split(":")
			csip = serverdata[0]
			csport = int(serverdata[1])
			pathhex = serverdata[2]
			pleng = len(pathhex)>>1
			pathlist = []
			for x in range(pleng):
				pathlist.append(int(pathhex[2*x:2*x+2],16))
			dout = struct.pack(">L%uB" % pleng,pleng,*pathlist)
			conn = MFSConn(csip,csport)
			data,length = conn.command(ANTOCS_CLEAR_ERRORS,CSTOAN_CLEAR_ERRORS,dout)
			del conn
			if length==1:
				res = (struct.unpack(">B",data))[0]
				if res>0:
					cmd_success = 1
		except Exception:
			tracedata = traceback.format_exc()
		url = fields.createrawlink({"CSclearerrors":""})
	if cmd_success==1:
		print("Status: 302 Found")
		print("Location: %s" % url)
		print("Content-Type: text/html; charset=UTF-8\r\n\r")
		# print("")
		print("""<!DOCTYPE html>""")
		print("""<html lang="en">""")
		print("""<head>""")
		print("""<meta http-equiv="Content-Type" content="text/html; charset=UTF-8" />""")
		print("""<meta http-equiv="Refresh" content="0; url=%s" />""" % url.replace('&','&amp;'))
		print("""<title>%s</title>""" % html_title)
		# leave this script a the beginning to prevent screen blinking when using dark mode
		print("""<script type="text/javascript"><!--//--><![CDATA[//><!--
			if (localStorage.getItem('theme')===null || localStorage.getItem('theme')==='dark') { document.documentElement.setAttribute('data-theme', 'dark');}	
			//--><!]]></script>""")		
		print("""<link rel="stylesheet" href="assets/mfs.css" type="text/css" />""")
		print("""</head>""")
		print("""<body>""")
		print("""<h1 align="center"><a href="%s">If you see this then it means that redirection didn't work, so click here</a></h1>""" % url)
		print("""</body>""")
		print("""</html>""")
		sys.exit(0)
	elif cmd_success==0:
		print("Content-Type: text/html; charset=UTF-8\r\n\r")
		print("""<!DOCTYPE html>""")
		print("""<html lang="en">""")
		print("""<head>""")
		print("""<meta http-equiv="Content-Type" content="text/html; charset=UTF-8" />""")
		print("""<meta http-equiv="Refresh" content="5; url=%s" />""" % url.replace('&','&amp;'))
		print("""<title>%s</title>""" % html_title)
		# leave this script a the beginning to prevent screen blinking when using dark mode
		print("""<script type="text/javascript"><!--//--><![CDATA[//><!--
			if (localStorage.getItem('theme')===null || localStorage.getItem('theme')==='dark') { document.documentElement.setAttribute('data-theme', 'dark');}	
			//--><!]]></script>""")		
		print("""<link rel="stylesheet" href="assets/mfs.css" type="text/css" />""")
		print("""</head>""")
		print("""<body>""")
		print("""<h3 align="center">Can't perform command - wait 5 seconds for refresh</h3>""")
		if tracedata:
			print("""<hr />""")
			print("""<pre>%s</pre>""" % tracedata)
		print("""</body>""")
		print("""</html>""")
		sys.exit(0)
		
